// Copyright (c) 2020, Andreas Schneider
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the <organization> nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package main

import (
	crand "crypto/rand"
	"math"
	"math/big"
	"math/rand"

	"github.com/alecthomas/kong"
)

type config struct {
	DataDirectory string `name:"data-dir" default:"data" help:"Directory to store all files in." env:"DATA_DIR"`
	Db            string `name:"db" default:"ShareDAV.db" help:"Database file to use." env:"DATABASE"`
}

type app struct {
	config
	CmdUser  CmdUser  `cmd:"" name:"user" help:"Manage users."`
	CmdShare CmdShare `cmd:"" name:"share" help:"Manage shares."`
	CmdServe CmdServe `cmd:"" name:"serve" help:"Serve HTTP requests." default:"1"`

	userStore  UserStore
	shareStore ShareStore
}

func main() {
	// When we need simple randomization (not cryptographically secure one), we should at least start
	// with some non-default-seed.
	seed, err := crand.Int(crand.Reader, big.NewInt(math.MaxInt64))
	if err != nil {
		panic("cannot initialize rand: " + err.Error())
	}
	rand.Seed(seed.Int64())

	var app app
	ctx := kong.Parse(&app)

	store, err := NewDBStore(app.config.Db)
	ctx.FatalIfErrorf(err)
	defer store.Close()

	app.userStore = store
	app.shareStore = store

	ctx.FatalIfErrorf(ctx.Run(&app))
}
